
/*
      _______                          .__ .__   .__   .__                 
      \      \    ____   __ __ _______ |__||  |  |  |  |__|  ____    ____  
      /   |   \ _/ __ \ |  |  \\_  __ \|  ||  |  |  |  |  | /  _ \  /    \ 
     /    |    \\  ___/ |  |  / |  | \/|  ||  |__|  |__|  |(  <_> )|   |  \
     \____|__  / \___  >|____/  |__|   |__||____/|____/|__| \____/ |___|  /
    =========\/======\/=================================================\/==
  v0.02 08/JAN/2008   Copyright 2007-2008 Scott D. Yelich SOME RIGHTS RESERVED
  ,-*~'`^`'~*-,._.,-*~'`^`'~*-,._.,-*~'`^`'~*-,._.-*~'`^`'~*-,._.,-*~'`^`'~*-,


  LICENSE:  Creative Commons Attribution 3.0 License.
  SEE:      http://creativecommons.org/licenses/by/3.0/


  Wed Jun  6 11:19:52 EDT 2007, v0.01 sdy
  Tue Jan  8 22:53:34 EST 2008, v0.02 sdy

*/

/*
    This is Socket.cpp -- part of Socket:

    The idea is to make a small and simple (to use) C++ wrapper to TCP
    sockets.  That is, this implementation is not trying to cover all
    uses of sockets such as UDP or multicast, etc.  The intended use of
    this class is to allow for a simple TCP server and a simple TCP
    client to communicate with each other (using sockets).

    This code doesn't do a lot of error checking/handling -- and what
    little it does do is C style -- mostly due to the fact that C++
    doesn't yet have a standard socket API and so the core system socket
    API is used -- and, of course, this is a C based API. A future
    version will use exceptions.


    Along with the standard ctor/dtor, this class implements an object
    that has the following components/features:

    "create()"  creates the socket.
    "bind(port)"  binds the socket to the specified port.
    "connect(host,port)"  connects to "host" on the specified port.
    "disconnect()"  does a shutdown{2} and then close on the socket.

    "send(string)"  sends the string through the socket (blocking).
    "recv_peek()"  does a select with 0 delay to see if there is input.
    "recv()"  receives input (blocking).

    "listen(port)"  sets the socket to list on the specified port.
    "reuseport()"  sets flag to allow reuse of local bind port.
    "accept()"  upon read ready on listen port, accept an incoming connection.

    "set_non_blocking(bool)"  sets flag to put socket in non-blocking mode.

    "setfd(int)"  sets the private socket descriptor value.
    "getfd()"  gets the private socket descriptor value.

    "input()"  returns a string with the value of the socket input buffer.
    "clearinput(int)"  clears the socket input buffer for "int" bytes length.


    EXAMPLE USAGE:

    #include "Socket.h"

    int port = 4322;
    std::string str;

    socket.create();

    //  reuseport is optional
    socket.reuseport();

    // (bind is optional)
    local_port =  1024+ (int)(30000 * (rand() / (RAND_MAX +1.0)));
    socket.bind(local_port);

    socket.connect("localhost", port);

    socket.send(str);

    socket.recv();  //  blocks
    str = socket.input();
    socket.clearinput();

*/


#include "Socket.h"


Socket::Socket()
  : socket_handle(0)
{
};

Socket::~Socket()
{
};

void
Socket::create()
{
  memset(&socket_addr, 0, sizeof(socket_addr));
  if ((socket_handle = socket(AF_INET, SOCK_STREAM, 0)) == -1 ) {
    throw TypedException<Socket::Errors>(Socket::Create, "create");
  }
}


void
Socket::bind(const int port)
{
  socket_addr.sin_family = AF_INET;
  socket_addr.sin_addr.s_addr = INADDR_ANY; // all local ips...
  socket_addr.sin_port = htons(port);
  memset(socket_addr.sin_zero, 0, sizeof (socket_addr.sin_zero));
  if (::bind(socket_handle, (sockaddr*)&socket_addr, sizeof(socket_addr)) == -1) {
    throw TypedException<Socket::Errors>(Socket::Bind, "bind");
  }
}

void
Socket::reuseport()
{
  int yes = 1;  // lame
  if (setsockopt(socket_handle, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) == -1 ) {
    throw TypedException<Socket::Errors>(Socket::Opt, "opts");
  }
}

void
Socket::listen(const int port)
{
  if (::listen(socket_handle, MAX_CONNECT_QUEUE) == -1 ) {
    throw TypedException<Socket::Errors>(Socket::Opt, "listen");
  }
}

int
Socket::accept()
{
  socklen_t addr_len = sizeof(remote_addr);
  int newfd = ::accept(socket_handle, (sockaddr*) &remote_addr, &addr_len);
  if (-1 == newfd) {
    throw TypedException<Socket::Errors>(Socket::Accept, "accept");
  }
  return newfd;
}

void
Socket::connect(const std::string host, const int port)
{
  socket_addr.sin_family = AF_INET;
  socket_addr.sin_port = htons(port);
  if (inet_pton(AF_INET, host.c_str(), &socket_addr.sin_addr) == -1 ) {
    // ?
  }
//  if ( errno == EAFNOSUPPORT ) return -1; // ?
  if ( ::connect(socket_handle, (sockaddr*) &socket_addr, sizeof(socket_addr)) == -1 ) {
    throw TypedException<Socket::Errors>(Socket::Connect, "connect");
  }
}

void
Socket::disconnect()
{
  if (socket_handle) { // can't close 0?
    ::shutdown(socket_handle, 2); // EC?
    ::close(socket_handle);  // EC?
    socket_handle = 0;
  }
}

int
Socket::send(const std::string s)
{
  int status = 0;
  //  status = ::send(socket_handle, s.c_str(), s.size(), MSG_NOSIGNAL);
  status = ::send(socket_handle, s.c_str(), s.size(), 0);
  //  EC -1?
  return status;
}

int
Socket::recv()
{
  char buffer[MAX_RECV+1];
  memset(buffer, 0, MAX_RECV+1 );
  int len = ::recv(socket_handle, buffer, MAX_RECV, 0);
  if (-1 != len) {
    ibuf.append(buffer,len); // memleak?
  } else {
    if (EAGAIN == errno) {
      std::cerr << "!";
//      sleep(1); // ?
    } else {
      throw TypedException<Socket::Errors>(Socket::Receive, "receive");
    }
  }
  return len;
}

void
Socket::setfd(const int fd)
{
  socket_handle = fd;
}

int
Socket::getfd()
{
  return socket_handle;
}

std::string &
Socket::input()
{
  return ibuf;
}

void
Socket::clearinput(int len)
{
  if (len > 0) {
    ibuf.erase(0, len);
  } else {
    ibuf = "";
  }
}

void
Socket::set_non_blocking ( const bool b )
{
  if (socket_handle) {
    int opts = fcntl(socket_handle, F_GETFL);
    if (opts > -1) {
      opts = (b) ? (opts | O_NONBLOCK) : (opts & ~O_NONBLOCK);
      fcntl(socket_handle, F_SETFL,opts);
    }
  }
}

bool
Socket::recv_peek(int timeout_seconds)
{
  bool rc = false;
  if (socket_handle) {
    fd_set rfds;
    FD_SET(socket_handle, &rfds);
    timeval tv;
    tv.tv_sec = timeout_seconds;
    tv.tv_usec = 0;
    if (::select(socket_handle+1, &rfds, 0, 0, &tv) == -1) {
      throw TypedException<Socket::Errors>(Socket::Select, "select");
    }
    if (FD_ISSET(socket_handle, &rfds)) {
      rc = true;
    }
  }
  return rc;
}

bool
Socket::writeable()
{
  bool rc = false;
  if (socket_handle) {
    fd_set wfds;
    FD_SET(socket_handle, &wfds);
    timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = 0;
std::cout << "writeable? select::writeable(0)" << std::endl;
    if (::select(socket_handle+1, 0, &wfds, 0, &tv) == -1) {
      throw TypedException<Socket::Errors>(Socket::Select, "select");
    }
    if (FD_ISSET(socket_handle, &wfds)) {
      rc = true;
    }
  }
  return rc;
}

